# -*- coding: utf-8 -*-
"""
Helper functions for multiprocessing in snr generation
"""

import numpy as np
import bilby

from .njit_functions import noise_weighted_inner_product


def noise_weighted_inner_prod(params):
    """
    Probaility of detection of GW for the given sensitivity of the detectors

    Parameters
    ----------
    params : list
        list of parameters for the inner product calculation
        List contains: \n
        params[0] : float
            mass_1
        params[1] : float
            mass_2
        params[2] : float
            luminosity_distance
        params[3] : float
            theta_jn
        params[4] : float
            psi
        params[5] : float
            phase
        params[6] : float
            ra
        params[7] : float
            dec
        params[8] : float
            geocent_time
        params[9] : float
            a_1
        params[10] : float
            a_2
        params[11] : float
            tilt_1
        params[12] : float
            tilt_2
        params[13] : float
            phi_12
        params[14] : float
            phi_jl
        params[15] : float
            approximant
        params[16] : float
            f_min
        params[17] : float
            duration
        params[18] : float
            sampling_frequency
        params[19] : int
            index tracker
        psds_list[20] : list
            list of psds for each detector
        detector_list[21:] : list
            list of detectors

    Returns
    -------
    SNRs_list : list
        contains opt_snr for each detector and net_opt_snr
    params[19] : int
        index tracker
    """

    bilby.core.utils.logger.disabled = True
    np.random.seed(88170235)
    parameters = {
        "mass_1": params[0],
        "mass_2": params[1],
        "luminosity_distance": params[2],
        "theta_jn": params[3],
        "psi": params[4],
        "phase": params[5],
        "geocent_time": params[8],
        "ra": params[6],
        "dec": params[7],
        "a_1": params[9],
        "a_2": params[10],
        "tilt_1": params[11],
        "tilt_2": params[12],
        "phi_12": params[13],
        "phi_jl": params[14],
    }

    waveform_arguments = dict(
        waveform_approximant=params[15],
        reference_frequency=20.0,
        minimum_frequency=params[16],
    )

    waveform_generator = bilby.gw.WaveformGenerator(
        duration=params[17],
        sampling_frequency=params[18],
        frequency_domain_source_model=bilby.gw.source.lal_binary_black_hole,
        waveform_arguments=waveform_arguments,
    )
    polas = waveform_generator.frequency_domain_strain(parameters=parameters)

    # h = F+.h+ + Fx.hx
    # <h|h> = <h+,h+> + <hx,hx> + 2<h+,hx>
    # <h|h> = <h+,h+> + <hx,hx>, if h+ and hx are orthogonal
    hp_inner_hp_list = []
    hc_inner_hc_list = []
    list_of_detectors = params[21:].tolist()
    psds_objects = params[20]
    for det in list_of_detectors:

        # need to compute the inner product for
        p_array = psds_objects[det].get_power_spectral_density_array(waveform_generator.frequency_array)
        idx2 = (p_array != 0.0) & (p_array != np.inf)
        hp_inner_hp = noise_weighted_inner_product(
            polas["plus"][idx2],
            polas["plus"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )
        hc_inner_hc = noise_weighted_inner_product(
            polas["cross"][idx2],
            polas["cross"][idx2],
            p_array[idx2],
            waveform_generator.duration,
        )

        # might need to add these lines in the future for waveform with multiple harmonics and h+ and hx are not orthogonal
        # hp_inner_hc = bilby.gw.utils.noise_weighted_inner_product(
        #     polas["plus"][idx2],
        #     polas["cross"][idx2],
        #     p_array[idx2],
        #     waveform_generator.duration,
        # )

        hp_inner_hp_list.append(hp_inner_hp)
        hc_inner_hc_list.append(hc_inner_hc)

    return (hp_inner_hp_list, hc_inner_hc_list, params[19])


# -*- coding: utf-8 -*-
"""
Helper functions for gwsnr. All functions are njit compiled.
"""

import numpy as np
from numba import njit

Gamma = 0.5772156649015329
Pi = np.pi
MTSUN_SI = 4.925491025543576e-06

@njit
def findchirp_chirptime(m1, m2, fmin):
    """
    Time taken from f_min to f_lso (last stable orbit). 3.5PN in fourier phase considered.

    Parameters
    ----------
    m1 : `float`
        Mass of the first body in solar masses.
    m2 : `float`
        Mass of the second body in solar masses.
    fmin : `float`
        Lower frequency cutoff.

    Returns
    -------
    chirp_time : float
        Time taken from f_min to f_lso (last stable orbit frequency).
    """

    # variables used to compute chirp time
    m = m1 + m2
    eta = m1 * m2 / m / m
    c0T = c2T = c3T = c4T = c5T = c6T = c6LogT = c7T = 0.0

    c7T = Pi * (
        14809.0 * eta * eta / 378.0 - 75703.0 * eta / 756.0 - 15419335.0 / 127008.0
    )

    c6T = (
        Gamma * 6848.0 / 105.0
        - 10052469856691.0 / 23471078400.0
        + Pi * Pi * 128.0 / 3.0
        + eta * (3147553127.0 / 3048192.0 - Pi * Pi * 451.0 / 12.0)
        - eta * eta * 15211.0 / 1728.0
        + eta * eta * eta * 25565.0 / 1296.0
        + eta * eta * eta * 25565.0 / 1296.0
        + np.log(4.0) * 6848.0 / 105.0
    )
    c6LogT = 6848.0 / 105.0

    c5T = 13.0 * Pi * eta / 3.0 - 7729.0 * Pi / 252.0

    c4T = 3058673.0 / 508032.0 + eta * (5429.0 / 504.0 + eta * 617.0 / 72.0)
    c3T = -32.0 * Pi / 5.0
    c2T = 743.0 / 252.0 + eta * 11.0 / 3.0
    c0T = 5.0 * m * MTSUN_SI / (256.0 * eta)

    # This is the PN parameter v evaluated at the lower freq. cutoff
    xT = np.power(Pi * m * MTSUN_SI * fmin, 1.0 / 3.0)
    x2T = xT * xT
    x3T = xT * x2T
    x4T = x2T * x2T
    x5T = x2T * x3T
    x6T = x3T * x3T
    x7T = x3T * x4T
    x8T = x4T * x4T

    # Computes the chirp time as tC = t(v_low)
    # tC = t(v_low) - t(v_upper) would be more
    # correct, but the difference is negligble.
    return (
        c0T
        * (
            1
            + c2T * x2T
            + c3T * x3T
            + c4T * x4T
            + c5T * x5T
            + (c6T + c6LogT * np.log(xT)) * x6T
            + c7T * x7T
        )
        / x8T
    )

@njit
def einsum1(m,n):
    """
    Function to calculate einsum of two 3x1 vectors
    """
    ans = np.zeros((3,3))
    ans[0,0] = m[0]*n[0]
    ans[0,1] = m[0]*n[1]
    ans[0,2] = m[0]*n[2]
    ans[1,0] = m[1]*n[0]
    ans[1,1] = m[1]*n[1]
    ans[1,2] = m[1]*n[2]
    ans[2,0] = m[2]*n[0]
    ans[2,1] = m[2]*n[1]
    ans[2,2] = m[2]*n[2]
    return ans
@njit
def einsum2(m,n):
    """
    Function to calculate einsum of two 3x3 matrices
    """
    ans = m[0,0]*n[0,0] + m[0,1]*n[0,1] + m[0,2]*n[0,2] + m[1,0]*n[1,0] + m[1,1]*n[1,1] + m[1,2]*n[1,2] + m[2,0]*n[2,0] + m[2,1]*n[2,1] + m[2,2]*n[2,2]
    return ans

@njit
def gps_to_gmst(gps_time):
    """
    Function to convert gps time to greenwich mean sidereal time
    """
    slope = 7.292115855382993e-05
    time0 = 1126259642.413
    time = gps_time - time0
    return slope*time+36137.068361399164

@njit
def ra_dec_to_theta_phi(ra, dec, gmst):
    """
    Function to convert ra and dec to theta and phi
    """

    phi = ra - gmst
    theta = np.pi / 2.0 - dec
    return theta, phi

@njit
def get_polarization_tensor(ra, dec, time, psi, mode='plus'):
    """
    Function to calculate the polarization tensor
    """
    gmst = np.fmod(gps_to_gmst(time), 2 * np.pi)
    theta, phi = ra_dec_to_theta_phi(ra, dec, gmst)
    u = np.array([np.cos(phi) * np.cos(theta), np.cos(theta) * np.sin(phi), -np.sin(theta)])
    v = np.array([-np.sin(phi), np.cos(phi), 0])
    m = -u * np.sin(psi) - v * np.cos(psi)
    n = -u * np.cos(psi) + v * np.sin(psi)

    if mode == 'plus':
        return einsum1(m, m) - einsum1(n, n)
    elif mode == 'cross':
        return einsum1(m, n) + einsum1(n, m)

@njit
def antenna_response(ra, dec, time, psi, detector_tensor, mode='plus'):
    """
    Function to calculate the antenna response

    Parameters
    ----------
    ra : `float`
        Right ascension of the source in radians.
    dec : float
        Declination of the source in radians.
    time : `float`
        GPS time of the source.
    psi : `float`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the detector (3x3 matrix)
    mode : `str`
        Mode of the polarization. Default is 'plus'.

    Returns
    -------
    antenna_response: `float`
        Antenna response of the detector.
    """

    polarization_tensor = get_polarization_tensor(ra, dec, time, psi, mode=mode)
    return einsum2(detector_tensor, polarization_tensor)

@njit
def antenna_response_array(ra, dec, time, psi, detector_tensor):
    """
    Function to calculate the antenna response in array form.

    Parameters
    ----------
    ra : `numpy.ndarray`
        Right ascension of the source in radians.
    dec : `numpy.ndarray`
        Declination of the source in radians.
    time : `numpy.ndarray`
        GPS time of the source.
    psi : `numpy.ndarray`
        Polarization angle of the source.
    detector_tensor : array-like
        Detector tensor for the multiple detectors (nx3x3 matrix), where n is the number of detectors.

    Returns
    -------
    antenna_response: `numpy.ndarray`
        Antenna response of the detector. Shape is (n, len(ra)).
    """

    len_det = len(detector_tensor)
    len_param = len(ra) 
    Fp = np.zeros((len_det, len_param))
    Fc = np.zeros((len_det, len_param))

    for j in range(len_det):
        for i in range(len_param):
            Fp[j,i] = antenna_response(ra[i], dec[i], time[i], psi[i], detector_tensor[j], mode="plus")
            Fc[j,i] = antenna_response(ra[i], dec[i], time[i], psi[i], detector_tensor[j], mode="cross")

    return Fp, Fc

@njit
def noise_weighted_inner_product(
    signal1, signal2, psd, duration,
):
    """
    Noise weighted inner product of two time series data sets.

    Parameters
    ----------
    signal1: `numpy.ndarray` or `float`
        First series data set.
    signal2: `numpy.ndarray` or `float`
        Second series data set.
    psd: `numpy.ndarray` or `float`
        Power spectral density of the detector.
    duration: `float`
        Duration of the data.
    """

    nwip_arr = np.conj(signal1) * signal2 / psd
    return 4 / duration * np.sum(nwip_arr)

@njit
def get_interpolated_snr(mass_1, mass_2, luminosity_distance, theta_jn, psi, geocent_time, ra, dec, detector_tensor, snr_halfscaled, ratio_arr, mtot_arr):
    """
    Function to calculate the interpolated snr for a given set of parameters

    Parameters
    ----------
    mass_1 : `numpy.ndarray`
        Mass of the first body in solar masses.
    mass_2 : `numpy.ndarray`
        Mass of the second body in solar masses.
    luminosity_distance : `float`
        Luminosity distance to the source in Mpc.
    theta_jn : `numpy.ndarray`
        Angle between the total angular momentum and the line of sight to the source in radians.
    psi : `numpy.ndarray`
        Polarization angle of the source.
    geocent_time : `numpy.ndarray`
        GPS time of the source.
    ra : ``numpy.ndarray`
        Right ascension of the source in radians.
    dec : `numpy.ndarray`
        Declination of the source in radians.
    detector_tensor : array-like
        Detector tensor for the detector (3x3 matrix)
    snr_halfscaled : `numpy.ndarray`
        Array of snr_halfscaled coefficients for the detector.
    ratio_arr : `numpy.ndarray`
        Array of mass ratio values for the snr_halfscaled coefficients.
    mtot_arr : `numpy.ndarray`
        Array of total mass values for the snr_halfscaled coefficients.
    
    Returns
    -------
    snr : `float`
        snr of the detector.
    """

    size = len(mass_1)
    len_ = len(detector_tensor)
    mtot = mass_1 + mass_2
    ratio = mass_2 / mass_1
    # get array of antenna response
    Fp, Fc = antenna_response_array(ra, dec, geocent_time, psi, detector_tensor)

    Mc = ((mass_1 * mass_2) ** (3 / 5)) / ((mass_1 + mass_2) ** (1 / 5))
    A1 = Mc ** (5.0 / 6.0)
    ci_2 = np.cos(theta_jn) ** 2
    ci_param = ((1 + np.cos(theta_jn) ** 2) / 2) ** 2
    
    size = len(mass_1)
    snr_half_ = np.zeros((len_,size))
    d_eff = np.zeros((len_,size))
    snr = np.zeros((len_,size))
    # loop over the detectors
    for j in range(len_):
        # loop over the parameter points
        for i in range(size):
            snr_half_coeff = snr_halfscaled[j]
            snr_half_[j,i] = cubic_spline_interpolator2d(mtot[i], ratio[i], snr_half_coeff, mtot_arr, ratio_arr)
            d_eff[j,i] =luminosity_distance[i] / np.sqrt(
                    Fp[j,i]**2 * ci_param[i] + Fc[j,i]**2 * ci_2[i]
                )

    snr = snr_half_ * A1 / d_eff
    snr_effective = np.sqrt(np.sum(snr ** 2, axis=0))

    return snr, snr_effective

@njit
def cubic_spline_interpolator2d(xnew, ynew, coefficients, x, y):
    """
    Function to calculate the interpolated value of snr_halfscaled given the mass ratio (ynew) and total mass (xnew). This is based off 2D bicubic spline interpolation.

    Parameters
    ----------
    xnew : `float`
        Total mass of the binary.
    ynew : `float`
        Mass ratio of the binary.
    coefficients : `numpy.ndarray`
        Array of coefficients for the cubic spline interpolation.
    x : `numpy.ndarray`
        Array of total mass values for the coefficients.
    y : `numpy.ndarray`
        Array of mass ratio values for the coefficients.

    Returns
    -------
    result : `float`
        Interpolated value of snr_halfscaled.
    """

    len_y = len(y)
    # y_idx = np.searchsorted(y, ynew)

    y_idx = np.searchsorted(y, ynew) - 1 if ynew > y[0] else 0
    # result = cubic_spline_interpolator(xnew, coefficients[y_idx], x)

    if y_idx == 0:  # lower end point
        result = cubic_spline_interpolator(xnew, coefficients[0], x)
        # print("a")
    elif y_idx+1 == len_y:  # upper end point
        result = cubic_spline_interpolator(xnew, coefficients[-1], x)
        # print("b")
    else:
        y_idx1 = y_idx - 1
        y_idx2 = y_idx
        y_idx3 = y_idx + 1
        y_idx4 = y_idx + 2
        coeff_low, coeff_high = 4, 8
        # print("c")
        y1, y2, y3, y4 = y[y_idx1], y[y_idx2], y[y_idx3], y[y_idx4]
        z1 = cubic_spline_interpolator(xnew, coefficients[y_idx1], x)
        z2 = cubic_spline_interpolator(xnew, coefficients[y_idx2], x)
        z3 = cubic_spline_interpolator(xnew, coefficients[y_idx3], x)
        z4 = cubic_spline_interpolator(xnew, coefficients[y_idx4], x)

        coeff = coefficients_generator(y1, y2, y3, y4, z1, z2, z3, z4)
        matrixD = coeff[coeff_low:coeff_high]
        matrixB = np.array([ynew**3, ynew**2, ynew, 1])
        result = np.dot(matrixB, matrixD)

    return result

@njit
def cubic_spline_interpolator(xnew, coefficients, x):
    """
    Function to calculate the interpolated value of snr_halfscaled given the total mass (xnew). This is based off 1D cubic spline interpolation.

    Parameters
    ----------
    xnew : `float`
        Total mass of the binary.
    coefficients : `numpy.ndarray`
        Array of coefficients for the cubic spline interpolation.
    x : `numpy.ndarray`
        Array of total mass values for the coefficients.

    Returns
    -------
    result : `float`
        Interpolated value of snr_halfscaled.
    """
    # Handling extrapolation
    i = np.searchsorted(x, xnew) - 1 if xnew > x[0] else 0

    # Calculate the relative position within the interval
    dx = xnew - x[i]

    # Calculate the interpolated value
    # Cubic polynomial: a + b*dx + c*dx^2 + d*dx^3
    a, b, c, d = coefficients[:, i]
    result = d + c*dx + b*dx**2 + a*dx**3
    return result

@njit
def coefficients_generator(y1, y2, y3, y4, z1, z2, z3, z4):
    """
    Function to generate the coefficients for the cubic spline interpolation of fn(y)=z.
    """
    matrixA = np.array([
        [y1**3, y1**2, y1, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        [y2**3, y2**2, y2, 1, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, y2**3, y2**2, y2, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, y3**3, y3**2, y3, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, y3**3, y3**2, y3, 1],
        [0, 0, 0, 0, 0, 0, 0, 0, y4**3, y4**2, y4, 1],
        [3*y2**2, 2*y2, 1, 0, -3*y2**2, -2*y2, -1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 3*y3**2, 2*y3, 1, 0, -3*y3**2, -2*y3, -1, 0],
        [6*y2, 2, 0, 0, -6*y2, -2, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 6*y3, 2, 0, 0, -6*y3, -2, 0, 0],
        [6*y1, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 0, 6*y4, 2, 0, 0],
    ])
    matrixC = np.array([z1, z2, z2, z3, z3, z4, 0, 0, 0, 0, 0, 0])
    return np.dot(np.linalg.inv(matrixA), matrixC)




"""
GWSNR: Gravitational Wave Signal-to-Noise Ratio
"""

import warnings
warnings.filterwarnings("ignore", "Wswiglal-redir-stdio")

__author__ = 'hemanta_ph <hemantaphurailatpam@gmail.com>'

__version__ = "0.1.4"

from .gwsnr import GWSNR




#!/usr/bin/env python
from setuptools import setup, find_packages
setup(name='gwsnr',
      version='0.2.0',
      description='Fast SNR interpolator',
      author='Hemantakumar Phurailatpam',
      license="MIT",
      author_email='hemantaphurailatpam@gmail.com',
      url='https://github.com/hemantaph/gwsnr',
      packages=find_packages(),
      install_requires=[
        "setuptools>=61.1.0",
        "numpy>=1.18",
        "bilby>=1.0.2",
        "pycbc>=2.0.4",
        "scipy>=1.9.0",
        "tqdm>=4.64.0",
        "gwpy>=2.1.5",
      ]
     )



# -*- coding: utf-8 -*-
"""
Helper functions for gwsnr
"""

import os
import json
import pickle
import numpy as np
import bilby


class NumpyEncoder(json.JSONEncoder):
    """
    Custom JSON encoder for numpy data types. It converts numpy.ndarray objects (and any nested-list composition 
    that includes ndarray objects) into regular lists for JSON serialization. This is particularly useful when 
    serializing data structures that include numpy arrays.
    """
    
    def default(self, obj):
        # Check if the object is an instance of numpy.ndarray
        if isinstance(obj, np.ndarray):
            return obj.tolist()  # Convert ndarray to list
        # Fallback to default behavior for other types
        return super(NumpyEncoder, self).default(obj)

def dealing_with_psds(psds=None, isit_psd_file=False, ifos=None, f_min=20., sampling_frequency=2048.):
    """
    Function to deal with psds inputs and for creating bilby.gw.detector.PowerSpectralDensity objects.

    Parameters
    ----------
    psds : dict
        dictionary of psds. psds.keys()=detector names, psds.values()=psds file names or pycbc psd names
    isit_psd_file : `bool` or `dict`
        If set True, the given value of psds param should be of psds instead of asd. If asd, set isit_psd_file=False. Default is False. If dict, it should be of the form {'L1':True, 'H1':True, 'V1':True} and should have keys for all the detectors.
    ifos : `list` or `None`
        List of interferometer objects. Default is None. If None, bilby's default interferometer objects will be used.
    f_min : `float`
        Minimum frequency of the psds. Default is 20.
    sampling_frequency : `float`
        Sampling frequency of the psds. Default is 2048.

    Returns
    ----------
    psds_list : `list`
        list of bilby.gw.detector.PowerSpectralDensity objects
    """
    
    if not psds:
        # if psds is not given, choose bilby's default psds
        print("psds not given. Choosing bilby's default psds")
        psds = dict()
        psds["L1"] = "aLIGO_O4_high_asd.txt"
        psds["H1"] = "aLIGO_O4_high_asd.txt"
        psds["V1"] = "AdV_asd.txt"
        detector_list = list(psds.keys())
        isit_psd_file_dict = dict(L1=False, H1=False, V1=False)
        # for Fp, Fc calculation
        ifos = bilby.gw.detector.InterferometerList(detector_list)
    else:
        # given psds is a dictionary
        # dict keys are the detectors names
        # dict values are the psds file names
        try:
            detector_list = list(psds.keys())
        except ValueError as e:
            raise ValueError("psds must be a dictionary with keys as detector names (e.g.'L1' for Livingston etc) and values as str") from e
        
        if type(isit_psd_file) == bool:  # this means all psds are 
            isit_psd_file_dict = dict()
            for det in detector_list:
                isit_psd_file_dict[det] = isit_psd_file
        elif list(isit_psd_file.keys())==detector_list:
            pass
        else:
            raise ValueError(f"isit_psd_file must be a dictionary with keys: {detector_list} and values as bool")

        ifos = []
        len_ = len(detector_list)
        for i in range(len_):
            try:
                # if ifos is not None, then use the given ifos
                # ifos is a list of bilby.gw.detector.Interferometer 
                ifos.append(ifos[i])
            except:
                # if ifos is None, then generate bilby's default ifos with the given list of detectors
                ifos.append(
                    bilby.gw.detector.InterferometerList(
                        [detector_list[i]]
                    )[0]
                )

    # generate bilby's psd objects
    psds_list = []
    detector_tensor_list = []
    i = 0
    for det in detector_list:
        # either provided psd or what's available in bilby
        if type(psds[det]) == str and psds[det][-3:] == "txt":
            if isit_psd_file_dict[det]:
                psds_list.append(
                    bilby.gw.detector.PowerSpectralDensity(
                        psd_file=psds[det]
                    )
                )
            else:
                psds_list.append(
                    bilby.gw.detector.PowerSpectralDensity(
                        asd_file=psds[det]
                    )
                )
            detector_tensor_list.append(ifos[i].detector_tensor)
            i += 1
        elif type(psds[det]) == str:
            psds_list.append(power_spectral_density_pycbc(psds[det]), f_min, sampling_frequency)
        else:
            raise ValueError(
                        "the psds format is not recognised. The parameter psds dict should contain chosen detector names as keys and corresponding psds txt file name (or name from pycbc psd)as their values"
                    )

    return psds_list, detector_tensor_list, detector_list

def power_spectral_density_pycbc(psd, f_min=20., sampling_frequency=2048.):
    """
    psd array finder from pycbc

    Parameters
    ----------
    psd : str
        name of the psd
        e.g. 'aLIGOaLIGODesignSensitivityT1800044'
    f_min : float
        minimum frequency of the psd
        default: 20.
    sampling_frequency : float
        sampling frequency of the psd
        default: 2048.

    Returns
    -------
    psd_array : bilby.gw.detector.psd.PowerSpectralDensity object
    """
    import pycbc
    import pycbc.psd

    delta_f = 1.0 / 16.0
    flen = int(sampling_frequency / delta_f)
    low_frequency_cutoff = f_min
    psd_ = pycbc.psd.from_string(psd, flen, delta_f, low_frequency_cutoff)
    return bilby.gw.detector.PowerSpectralDensity(
        frequency_array=psd_.sample_frequencies, psd_array=psd_.data
    )

# interpolator check and generation
def interpolator_check(
        param_dict_given, interpolator_dir, create_new,
):
    """
    Function for interpolator (snr_halfsacaled) check and generation if not exists.

    Parameters
    ----------
    param_dict_given : dict
        dictionary of parameters based on which the existence of interpolator will be checked
    interpolator_dir : str
        path to the interpolator pickle file
    create_new : bool
        if True, new interpolator will be generated even if the interpolator exists
        if False, existing interpolator will be used if exists, otherwise new interpolator will be generated

    Returns
    ----------
    psds_list_ : list
        list of psd objects
    detector_tensor_list_ : list
        list of detector tensors
    detector_list_ : list
        list of detector names
    """

    detector_list = param_dict_given["detector"]
    psds_list = param_dict_given["psds"]
    detector_tensor_list = param_dict_given["detector_tensor"]
    detector_list_ = []
    detector_tensor_list_ = []
    psds_list_ = []
    path_interpolator_ = []

    # getting interpolator if exists
    k = 0
    for det in detector_list:
        param_dict_given["detector"] = det
        param_dict_given["psds"] = psds_list[k]
        param_dict_given["detector_tensor"] = detector_tensor_list[k]
        path_interpolator, it_exist = interpolator_pickle_path(
            param_dict_given, interpolator_dir
        )
        if create_new:
            it_exist = False
        if it_exist:
            print(
                f"Interpolator will be loaded for {det} detector from {path_interpolator}"
            )
        else:
            print(
                f"Interpolator will be generated for {det} detector at {path_interpolator}"
            )

            detector_list_.append(det)
            psds_list_.append(psds_list[k])
            detector_tensor_list_.append(detector_tensor_list[k])

        path_interpolator_.append(path_interpolator)
        k += 1

    return psds_list_, detector_tensor_list_, detector_list_, path_interpolator_

def interpolator_pickle_path(
    param_dict_given, path="./interpolator_pickle"
):
    """
    Function for storing or getting interpolator (snr_halfsacaled) pickle path

    Parameters
    ----------
    param_dict_given : dict
        dictionary of parameters based on which the existence of interpolator will be checked
    path : str
        path to the directory where the interpolator pickle file will be stored

    Returns
    -------
    path_interpolator : str
        path to the interpolator pickle file
        e.g. './interpolator_pickle/L1/halfSNR_dict_0.pickle'
    it_exist: bool
        True if the interpolator exists
        False if the interpolator does not exists
    """

    detector = param_dict_given["detector"]
    # arg no. for the detector from the detector list
    det_path = path + "/" + detector

    # check if path exists
    if not os.path.exists(path):
        os.makedirs(path)

    # check if detector path exists
    if not os.path.exists(det_path):
        os.makedirs(det_path)

    # check if param_dict_list.pickle exists
    if not os.path.exists(det_path + "/param_dict_list.pickle"):
        dict_list = []
        with open(det_path + "/param_dict_list.pickle", "wb") as handle:
            pickle.dump(dict_list, handle, protocol=pickle.HIGHEST_PROTOCOL)

    # for checking existing interpolator pickle in the det_path/param_dict_list.pickle file
    param_dict_stored = pickle.load(
        open(det_path + "/param_dict_list.pickle", "rb")
    )

    len_ = len(param_dict_stored)
    #del param_dict_given["psds"]
    param_dict_given["detector_tensor"] = str(param_dict_given["detector_tensor"])
    # print("\n\n", param_dict_given)
    # print("\n\n",param_dict_stored)
    if param_dict_given in param_dict_stored:
        # try and except is added so that user can regenerate a new interpolator pickle file just by
        # deleting the right file and reruing gwsnr with that params again
        # also, if the user delete the file by mistake, it will generate in the next run
        idx = param_dict_stored.index(param_dict_given)
        # check if interpolator pickle exists
        # get halfSNR interpolator if exists
        path_interpolator = det_path + "/halfSNR_dict_" + str(idx) + ".pickle"
        # there will be exception if the file is deleted by mistake
        if os.path.exists(path_interpolator):
            it_exist = True
        else:
            it_exist = False

    # if related dict not found in the param_dict_list.pickle
    else:
        it_exist = False
        path_interpolator = det_path + "/halfSNR_dict_" + str(len_) + ".pickle"
        # print("related dict not found in the param_dict_list.pickle, new interpolator will be generated")

        # store the pickle dict
        param_dict_stored.append(param_dict_given)
        with open(det_path + "/param_dict_list.pickle", "wb") as handle:
            pickle.dump(param_dict_stored, handle, protocol=pickle.HIGHEST_PROTOCOL)

    # print(f"In case if you need regeneration of interpolator of the given gwsnr param, please delete this file, {path_interpolator} \n")

    return (path_interpolator, it_exist)

def load_json(file_name):
    """Load a json file.

    Parameters
    ----------
    file_name : `str`
        json file name for storing the parameters.

    Returns
    ----------
    param : `dict`
    """
    with open(file_name, "rb") as handle:
        param = pickle.load(handle)

    return param

def save_json(param, file_name):
    """Save a json file.

    Parameters
    ----------
    param : `dict`
        dictionary of parameters.
    file_name : `str`
        json file name for storing the parameters.
    """
    with open(file_name, "wb") as handle:
        pickle.dump(param, handle, protocol=pickle.HIGHEST_PROTOCOL)

def save_json_dict(dict, file_name):
    """Save a json file.

    Parameters
    ----------
    param : `dict`
        dictionary of parameters.
    file_name : `str`
        json file name for storing the parameters.
    """

    json_dump = json.dumps(dict, cls=NumpyEncoder)
    with open(file_name, "w") as write_file:
        json.dump(json.loads(json_dump), write_file, indent=4)


def load_json_dict(file_name):
    """Load a json file.

    Parameters
    ----------
    file_name : `str`
        json file name for storing the parameters.

    Returns
    ----------
    param : `dict`
    """
    
    with open(file_name, "r", encoding="utf-8") as f:
        param = json.load(f)

    return param